/* -*- Mode: javascript; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ubufox.
 *
 * The Initial Developer of the Original Code is
 * Canonical Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Alexander Sack <asac@jwsdot.com> - Canonical Ltd.
 *   Arzhel Younsi <xionox@gmail.com>
 *   Chris Coulson <chris.coulson@canonical.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const { classes: Cc, interfaces: Ci, utils: Cu } = Components;

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
Cu.import("resource://gre/modules/FileUtils.jsm");
Cu.import("resource://gre/modules/ctypes.jsm");
Cu.import("resource://ubufox/libs/glib.jsm");
Cu.import("resource://ubufox/libs/gobject.jsm");
Cu.import("resource://ubufox/libs/gio.jsm");

const NS_XPCOM_CURRENT_PROCESS_DIR        = "XCurProcD";
const NS_GRE_DIR                          = "GreD";
const NS_USER_PLUGINS_DIR                 = "UserPlugins";
const NS_APP_PLUGINS_DIR                  = "APlugns";
const NS_SYSTEM_PLUGINS_DIR               = "SysPlugins";
const XRE_SYS_LOCAL_EXTENSION_PARENT_DIR  = "XRESysLExtPD";
const XRE_SYS_SHARE_EXTENSION_PARENT_DIR  = "XRESysSExtPD";
const XRE_EXTENSIONS_DIR_LIST             = "XREExtDL";

["LOG", "WARN", "ERROR"].forEach(function(aName) {
  this.__defineGetter__(aName, function() {
    Components.utils.import("resource://gre/modules/AddonLogging.jsm");

    LogManager.getLogger("ubufox.urn", this);
    return this[aName];
  });
}, this);

function UpdateNotifier() {
  LOG("Starting");
  this.wrappedJSObject = this;

  var listeners = [];
  var watches = {};
  var timers = {};
  var dispatched = {};
  var plugins = {};

  /**
   * Add a listener to receive update notifications
   * Calling this inside a notification may produce undefined results
   *
   * @param  aListener
   *         The listener object
   */
  this.addListener = function UN_addListener(aListener) {
    if (!aListener) {
      throw new Error("A listener must be specified");
    }

    if (typeof(aListener) != "object") {
      throw new Error("Invalid listener");
    }

    if (listeners.indexOf(aListener) != -1) {
      throw new Error("Trying to register an observer more than once");
    }

    listeners.push(aListener);

    for (let event in dispatched) {
      try {
        if (event in aListener) {
          aListener[event].apply(aListener, dispatched[event]);
        }
      } catch(e) {
        Cu.reportError(e);
      }
    }
  };

  /**
   * Remove a listener and don't receive any more notifications.
   * Calling this inside a notification may produce undefined results
   *
   * @param  aListener
   *         The listener to remove
   */
  this.removeListener = function UN_removeListener(aListener) {
    let index = listeners.indexOf(aListener);
    if (index == -1) {
      throw new Error("Trying to remove an observer that was never registered");
    }

    listeners.splice(index, 1);
  };

  function dispatchNotification() {
    let event = arguments[0];
    let args = Array.prototype.slice.call(arguments, 1);
    dispatched[event] = args;

    listeners.forEach(function(listener) {
      try {
        if (event in listener) {
          listener[event].apply(listener, args);
        }
      } catch(e) {
        Cu.reportError(e);
      }
    });
  }

  function handleAppUpdated() {
    if (!("onAppUpdated" in dispatched)) {
      dispatchNotification("onAppUpdated");
    }
  }

  function handlePluginsChanged() {
    let ph = Cc["@mozilla.org/plugin/host;1"].getService(Ci.nsIPluginHost);

    try {
      ph.reloadPlugins(false);
    } catch(e) { return; }

    let oldPluginNamesByPath = {};
    for (let name in plugins) {
      oldPluginNamesByPath[plugins[name].path] = name;
    }

    // Here's how this works:
    // - We iterate over the updated list of plugin tags
    // - For each plugin tag, we check if we have old data for that plugin,
    //   first looking up by name and then by path (we use name in case
    //   the install location changed but fall back to path in case the new
    //   name is different. Obviously, we won't identify the old plugin data
    //   if the update has a new path and name - in this case, we will mistake
    //   it as a new install. Oh well...)
    // - We keep a count of how many plugin tags there are for each plugin.
    // - For each plugin, we compare the newest mtime with the recorded mtime
    //   from the old plugin data (if it exists). If the old data exists and the
    //   mtime is different, then the plugin was updated
    // - If there is more than one plugin tag for a plugin, this means that the
    //   old version of a plugin might still be in use. In this case, we
    //   require a restart. If there is only one plugin tag for an updated
    //   plugin, then the old version is definitely not in use and the next
    //   use of this plugin will pick up the new version without requiring
    //   a restart
    let seenPlugins = {};
    for each (let pluginTag in ph.getPluginTags()) {
      let oldPluginData;
      if (!(pluginTag.name in plugins)) {
        oldPluginData = plugins[oldPluginNamesByPath[pluginTag.fullpath]];
      } else {
        oldPluginData = plugins[pluginTag.name];
      }

      if (!(pluginTag.name in seenPlugins)) {
        seenPlugins[pluginTag.name] = {
          "pluginData": {
            "path": pluginTag.fullpath,
            "lastModifiedTime": (new FileUtils.File(pluginTag.fullpath))
                                              .lastModifiedTime
          },
          "count": 1,
          "oldPluginData": oldPluginData
        }
      } else {
        ++seenPlugins[pluginTag.name].count;
        if (pluginTag.fullpath !=
            seenPlugins[pluginTag.name].pluginData.path) {
          let mtime = (new FileUtils.File(pluginTag.fullpath))
                                    .lastModifiedTime;
          if (mtime > seenPlugins[pluginTag.name].pluginData
                                                 .lastModifiedTime) {
            seenPlugins[pluginTag.name].pluginData.path = pluginTag.fullpath;
            seenPlugins[pluginTag.name].pluginData.lastModifiedTime = mtime;
          }
        }
      }
    }

    let foundUpdated = false;
    let requiresRestart = false;
    for (let name in seenPlugins) {
      if (!foundUpdated && seenPlugins[name].oldPluginData &&
          seenPlugins[name].oldPluginData.lastModifiedTime !=
          seenPlugins[name].pluginData.lastModifiedTime) {
        foundUpdated = true;
      }

      if (!requiresRestart && seenPlugins[name].count > 1) {
        requiresRestart = true;
      }

      plugins[name] = seenPlugins[name].pluginData;
    }

    if (foundUpdated) {
      if ("onPluginsUpdated" in dispatched && 
          dispatched["onPluginsUpdated"][0]) {
        return;
      }
      dispatchNotification("onPluginsUpdated", requiresRestart);
    }
  }

  function handleAddonsChanged() {
    // We don't know how to handle addon changes yet
  }

  function handler(aFile, aEventType, aIsDir) {
    let cpath = new ctypes.char.ptr;
    try {
      if (aEventType == gio.G_FILE_MONITOR_EVENT_CHANGED ||
          aEventType == gio.G_FILE_MONITOR_EVENT_DELETED ||
          aEventType == gio.G_FILE_MONITOR_EVENT_CREATED ||
          aEventType == gio.G_FILE_MONITOR_EVENT_ATTRIBUTE_CHANGED) {
        cpath = gio.g_file_get_path(aFile);
        path = cpath.readString();
        var type;
        if (!(path in watches) && aIsDir) {
          type = watches[(new FileUtils.File(path)).parent.path].type;
        } else {
          type = watches[path].type;
        }

        if (!type) {
          Cu.reportError("Got event for a path not in our watch list: " +
                         path)
          return;
        }

        // Wait for the location to settle before dispatching the notification
        if (!(type in timers)) {
          timers[type] = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
          timers[type].initWithCallback(function() {
            delete timers[type];
            ((new Object({"app"    : handleAppUpdated,
                          "plugins": handlePluginsChanged,
                          "addons" : handleAddonsChanged}))[type])();
          }, 10000, Ci.nsITimer.TYPE_ONE_SHOT);
        } else {
          timers[type].delay = 10000;
        }
      }
    } finally {
      glib.g_free(cpath);
    }
  }

  function watchOneFile(aFile, aType) {
    if (!aFile.exists() || !aFile.isFile()) {
      return;
    }

    aFile.normalize();

    let file = gio.g_file_new_for_path(aFile.path);
    if (file.isNull()) {
      Cu.reportError("OOM: Failed to get GFile for path " + aFile.path);
      return;
    }

    let mon = gio.g_file_monitor_file(file, gio.G_FILE_MONITOR_NONE,
                                      null, null);
    gobject.g_object_unref(file);
    if (mon.isNull()) {
      Cu.reportError("Failed to create monitor for path " + aFile.path);
      return;
    }

    let id = gobject.GSignalConnect(mon, "changed", function(aMonitor,
                                                             aFile,
                                                             aOtherFile,
                                                             aEventType,
                                                             aUserData) {
      handler(aFile, aEventType, false);    
    }, ctypes.void_t, [gio.GFileMonitor.ptr, gio.GFile.ptr, gio.GFile.ptr,
                       gio.GFileMonitorEvent, glib.gpointer]);
    watches[aFile.path] = {monitor: mon, sig_id: id, type: aType};
    LOG("Watching " + aFile.path);
  }

  function watchOneDir(aKey, aPaths, aType) {
    let nsifile;
    if (aKey instanceof Ci.nsIFile) {
      nsifile = aKey;
      aPaths.forEach(function(path) {
        nsifile.append(path);
      });
    } else {
      nsifile = FileUtils.getDir(aKey, aPaths);
    }
    if (!nsifile.exists() || !nsifile.isDirectory()) {
      return;
    }

    nsifile.normalize();

    if (nsifile.path in watches) {
      return;
    }

    let file = gio.g_file_new_for_path(nsifile.path);
    if (file.isNull()) {
      Cu.reportError("OOM: Failed to get GFile for path " + nsifile.path);
      return;
    }

    let mon = gio.g_file_monitor_directory(file, gio.G_FILE_MONITOR_NONE,
                                           null, null);
    gobject.g_object_unref(file);
    if (mon.isNull()) {
      Cu.reportError("Failed to create monitor for path " + nsifile.path);
      return;
    }

    let id = gobject.GSignalConnect(mon, "changed", function(aMonitor,
                                                             aFile,
                                                             aOtherFile,
                                                             aEventType,
                                                             aUserData) {
      handler(aFile, aEventType, true);
    }, ctypes.void_t, [gio.GFileMonitor.ptr, gio.GFile.ptr, gio.GFile.ptr,
                       gio.GFileMonitorEvent, glib.gpointer]);
    watches[nsifile.path] = {monitor: mon, sig_id: id, type: aType};
    LOG("Watching " + nsifile.path);
  }

  function watchList(aKey, aType, aFilterList) {
    let e = Services.dirsvc.get(aKey, Ci.nsISimpleEnumerator);
    let files = [];
    while (e.hasMoreElements()) {
      files.push(e.getNext().QueryInterface(Ci.nsIFile));
    }

    if (aFilterList) {
      files = files.filter(function(aValue, aIndex, aArray) {
        if (aFilterList.indexOf(aValue.parent.path) != -1) {
          return true;
        }

        return false;
      });
    }

    files.forEach(function(file) {
      if (!(file.exists())) {
        return;
      }

      if (file.isDirectory()) {
        watchOneDir(file, [], aType);
      } else {
        watchOneFile(file, aType);
      }
    });
  }

  Services.obs.addObserver(function(aSubject, aTopic, aData) {
    if (aSubject == "xpcom-shutdown") {
      Services.obs.removeObserver(arguments.callee, "xpcom-shutdown", false);
      for (let path in watches) {
        gobject.GSignalHandlerDisconnect(watches[path].monitor,
                                         watches[path].sig_id);
        gobject.g_object_unref(watches[path].monitor);
      }
    }
  }, "xpcom-shutdown", false);

  var idle = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
  idle.initWithCallback(function() {
    let kungFuDeathGrip = idle;
    watchOneDir(NS_XPCOM_CURRENT_PROCESS_DIR, [], "app");
    watchOneDir(NS_GRE_DIR, [], "app");
    watchOneDir(NS_APP_PLUGINS_DIR, [], "plugins");
    watchOneDir(NS_SYSTEM_PLUGINS_DIR, [], "plugins");
    //watchOneDir(XRE_SYS_LOCAL_EXTENSION_PARENT_DIR,
    //            [Services.appinfo.ID], "addons");
    //watchOneDir(XRE_SYS_SHARE_EXTENSION_PARENT_DIR,
    //            [Services.appinfo.ID], "addons");
    //watchOneDir(NS_XPCOM_CURRENT_PROCESS_DIR,
    //            ["extensions"], "addons");
    //watchList(XRE_EXTENSIONS_DIR_LIST, "addons",
    //          [FileUtils.getDir(XRE_SYS_LOCAL_EXTENSION_PARENT_DIR,
    //                            [Services.appinfo.ID]).path,
    //           FileUtils.getDir(XRE_SYS_SHARE_EXTENSION_PARENT_DIR,
    //                            [Services.appinfo.ID]).path,
    //           FileUtils.getDir(NS_XPCOM_CURRENT_PROCESS_DIR,
    //                            ["extensions"]).path]);

    idle.initWithCallback(function() {
      let kungFuDeathGrip = idle;
      // Would use the AddonManager here, but it doesn't expose full paths.
      // We use a delay here to avoid starting the plugin host at startup
      // if it isn't otherwise needed
      Cc["@mozilla.org/plugin/host;1"].getService(Ci.nsIPluginHost)
                                      .getPluginTags()
                                      .forEach(function(aPluginTag) {
        plugins[aPluginTag.name] = {
          "path": aPluginTag.fullpath,
          "lastModifiedTime": (new FileUtils.File(aPluginTag.fullpath))
                                            .lastModifiedTime
        };
      });
    }, 8000, Ci.nsITimer.TYPE_ONE_SHOT);

  }, 0, Ci.nsITimer.TYPE_ONE_SHOT);
}

UpdateNotifier.prototype = {
  classDescription: "Update Notifier",
  classID: Components.ID("{799a4700-2085-40e8-853a-05dcf8e95e41}"),
  contractID: "@ubuntu.com/update-notifier;1",
  QueryInterface: XPCOMUtils.generateQI([Ci.nsISupports])
};

const NSGetFactory = XPCOMUtils.generateNSGetFactory([UpdateNotifier]);
